// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QD multiplication (dd * dd)
//
// Calling Sequence :
// a * b
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function c = %dd_m_dd(a,b)
  [m1,n1] = size(a.hi);
  [m2,n2] = size(b.hi);
  
  t = 134217729 * a.hi;
  ah = t - (t - a.hi);
  al = a.hi - ah;
    
  t = 134217729 * b.hi;
  bh = t - (t - b.hi);
  bl = b.hi - bh;
  
  //(1~1)~(1~1),(1~1)~(m~n),(m~n)~(1~1)
  if ((m1 == 1 & n1 == 1) | (m2 == 1 & n2 == 1)) then
    p1 = a.hi * b.hi;
    p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
    p2 = p2 + a.hi * b.lo;
    p2 = p2 + a.lo * b.hi;
    
    tmp = p1 + p2;
    c = tlist(['dd','hi','lo'],tmp,p2-(tmp-p1));
  //(m~n)~(n~1)
  elseif n1 == m2 & n2 == 1 then
    p1 = a.hi(:,1) * b.hi(1,1);
    p2 = ((ah(:,1) * bh(1,1) - p1) + ah(:,1) * bl(1,1) + al(:,1) * bh(1,1)) + al(:,1) * bl(1,1);
    p2 = p2 + a.hi(:,1) * b.lo(1,1);
    p2 = p2 + a.lo(:,1) * b.hi(1,1);
      
    chi(:,1) = p1 + p2;
    clo(:,1) = p2-(chi(:,1)-p1);
    
    for k=2:n1
      p1 = a.hi(:,k) * b.hi(k,1);
      p2 = ((ah(:,k) * bh(k,1) - p1) + ah(:,k) * bl(k,1) + al(:,k) * bh(k,1)) + al(:,k) * bl(k,1);
      p2 = p2 + a.hi(:,k) * b.lo(k,1);
      p2 = p2 + a.lo(:,k) * b.hi(k,1);
      
      chi(:,k) = p1 + p2;
      clo(:,k) = p2-(chi(:,k)-p1);
      
      sh = chi(:,k-1) + chi(:,k);
      v = sh - chi(:,k-1);
      eh = (chi(:,k-1) - (sh - v)) + (chi(:,k) - v);
      
      sl = clo(:,k-1) + clo(:,k);
      eh = eh + sl;
      chi(:,k) = sh + eh;
      clo(:,k) = eh - (chi(:,k) - sh);
    end
    c = tlist(['dd','hi','lo'],chi(:,n1),clo(:,n1));
  //(1~l)~(l~n)
  elseif m1 == 1 & n2 == m2 then
    p1 = a.hi(1,1) * b.hi(1,:);
    p2 = ((ah(1,1) * bh(1,:) - p1) + ah(1,1) * bl(1,:) + al(1,1) * bh(1,:)) + al(1,1) * bl(1,:);
    p2 = p2 + a.hi(1,1) * b.lo(1,:);
    p2 = p2 + a.lo(1,1) * b.hi(1,:);
      
    chi(1,:) = p1 + p2;
    clo(1,:) = p2-(chi(1,:)-p1);
    
    for k=2:m2
      p1 = a.hi(1,k) * b.hi(k,:);
      p2 = ((ah(1,k) * bh(k,:) - p1) + ah(1,k) * bl(k,:) + al(1,k) * bh(k,:)) + al(1,k) * bl(k,:);
      p2 = p2 + a.hi(1,k) * b.lo(k,:);
      p2 = p2 + a.lo(1,k) * b.hi(k,:);
      
      chi(k,:) = p1 + p2;
      clo(k,:) = p2-(chi(k,:)-p1);
      
      sh = chi(k-1,:) + chi(k,:);
      v = sh - chi(k-1,:);
      eh = (chi(k-1,:) - (sh - v)) + (chi(k,:) - v);
      
      sl = clo(k-1,:) + clo(k,:);
      eh = eh + sl;      
      chi(k,:) = sh + eh;
      clo(k,:) = eh - (chi(k,:) - sh);
    end
    c = tlist(['dd','hi','lo'],chi(m2,:),clo(m2,:));
  //(m~l)~(l~n)
  elseif n1 == m2 then
    C = zeros(m1,n2);
    c = tlist(['dd','hi','lo'],C,C);
    for i=1:m1
      p1 = a.hi(i,1) * b.hi(1,:);
        p2 = ((ah(i,1) * bh(1,:) - p1) + ah(i,1) * bl(1,:) + al(i,1) * bh(1,:)) + al(i,1) * bl(1,:);
        p2 = p2 + a.hi(i,1) * b.lo(1,:);
        p2 = p2 + a.lo(i,1) * b.hi(1,:);
    
        chi(1,:) = p1 + p2;
        clo(1,:) = p2-(chi(1,:)-p1);
      for k=2:n1
        p1 = a.hi(i,k) * b.hi(k,:);
        p2 = ((ah(i,k) * bh(k,:) - p1) + ah(i,k) * bl(k,:) + al(i,k) * bh(k,:)) + al(i,k) * bl(k,:);
        p2 = p2 + a.hi(i,k) * b.lo(k,:);
        p2 = p2 + a.lo(i,k) * b.hi(k,:);
    
        chi(k,:) = p1 + p2;
        clo(k,:) = p2-(chi(k,:)-p1);
        
        sh = chi(k-1,:) + chi(k,:);
        v = sh - chi(k-1,:);
        eh = (chi(k-1,:) - (sh - v)) + (chi(k,:) - v);
      
        sl = clo(k-1,:) + clo(k,:);    
        eh = eh + sl;
        chi(k,:) = sh + eh;
        clo(k,:) = eh - (chi(k,:) - sh);
      end
      c(i,:) = dd(chi(n1,:),clo(n1,:));
    end
  else
    error(10) //Inconsistent multiplication.
    abort;
  end
endfunction

