// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QD multiplication (dd * qd)
//
// Calling Sequence :
// a * b
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function c = %dd_m_qd(a,b)
  [m1,n1] = size(a.hi);
  [m2,n2] = size(b.d);
  //(1~1)~(1~1),(1~1)~(m~n),(m~n)~(1~1)
  if (m1 == 1 & n1 == 1) | (m2 == 1 & n2 == 1) then
    c = prod_qd_dd(b,a)
  //(m~n)~(n~1)
  elseif n1 == m2 & n2 == 1 then
    A = a';B = b';
    c = prod_qd_dd(B(1,1),A(1,:));
    for k=2:n1
      d = prod_qd_dd(B(1,k),A(k,:));
      c = %qd_a_qd(c,d);
    end
    c = c';
  //(1~l)~(l~n)
  elseif m1 == 1 & n2 == m2 then
      A = a';B = b';
    c = prod_qd_dd(B(:,1),A(1,1));
    for k=2:n1
      d = prod_qd_dd(B(:,k),A(k,1));
      c = %qd_a_qd(c,d);
    end
    c = c';
  //(m~l)~(l~n)
  elseif n1 == m2 then
    A = a'; B=b';
    c = %qd_m_dd(B,A);
    c = c';
  else
    error(10) //Inconsistent multiplication.
    abort;
  end
endfunction



function c = prod_qd_dd(a,b)
  //QD multiplication
  //QD * DD
  [p1,e1] = two_prod(a(2),b.hi);
  [p2,e2] = two_prod(a(2),b.lo);
  [p3,e3] = two_prod(a(3),b.hi);
  [p4,e4] = two_prod(a(3),b.lo);
  [p5,e5] = two_prod(a(4),b.hi);
  //O() terms
  [p2,p3,e1] = three_sum(p2,p3,e1);
  //O(^2) terms
  //five term sum
  [p3,p4,p5] = three_sum(p3,p4,p5);
  [e2,e3] = two_sum(e2,e3);
  [p3,e2] = two_sum(p3,e2);
  [p4,e3] = two_sum(p4,e3);
  [p4,e2] = two_sum(p4,e2);
  p5 = p5+e2+e3;
  //O(^3) terms
  s = a(4)*b.lo + a(5)*b.hi + e4 + e5;
  [p4,e1] = three_sum2(p4,e1,s);
  p5 = p5+e1;
  [t1,t2,t3,t4] = renormalize(p1,p2,p3,p4,p5);
  c = tlist(['qd','d'],t1,t2,t3,t4);
endfunction
