// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// Vector norm of DD number
//
// Calling Sequence :
// norm(x,N)
//
// Arguments :
// x : DD vector
// N : 1 or 2 or %inf
//
// ***TODO : apply for matrix
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function z = %dd_norm(x,N)
  //vector norm
  [m,n] = size(x.hi);
  if n ~= 1 then
    disp('Error : input number is not vector.');
  else
  z = tlist(['dd','hi','lo'],0,0);
  select N
    case 1 //1-norm
      for i=1:m
        z = z + abs(x(i,1));
      end
    case 2 //2-norm
      t = 134217729 * x.hi;
      ah = t - (t - x.hi);
      al = x.hi - ah;
      
      p1 = x.hi .* x.hi;
      p2 = ((ah .* ah - p1) + 2.0 * (ah .* al)) + al .* al;
      p2 = p2 + 2.0 * (x.hi .* x.lo);
      
      chi = p1 + p2;
      clo = p2 - (chi - p1);
      
      for k=2:m 
        sh = chi(k-1) + chi(k);
        v = sh - chi(k-1);
        eh = (chi(k-1) - (sh - v)) + (chi(k) - v);
        
        sl = clo(k-1) + clo(k);
        eh = eh + sl;
        chi(k) = sh + eh;
        clo(k) = eh - (chi(k) - sh);
      end
      
      c = sqrt(chi(m)); // initial value   
      p1 = c*c;
      t = 134217729 * c;
      ah = t - (t - c);
      al = c - ah;
      p2 = ((ah * ah - p1) + 2.0 * ah * al) + al * al;
      cc = (chi(m) - p1 - p2 + clo(m)) / (c * 2.0); // Newton Iterative
      
      tmp =  c + cc;
      z = tlist(['dd','hi','lo'],tmp,cc-(tmp-c));
    case %inf //-norm
      z = abs(x);
      x_max = z(1)
      for i=2:m
        if(z(i) > x_max)
          x_max = z(i);
        end
      end
    else
      disp('n is not 1 or 2 or %inf')
    end
  end
endfunction

