// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// LU decomposition
//
// Calling Sequence :
// lu(A)
//
// Arguments :
// A : QD matrix
//
// Return :
// L : (n,n) lower triangular matrix
// U : (n,n) upper triangular matrix
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function [L,U] = %qd_lu(A)
  [m,n] = size(A.d)
  if (m~=n) then
    disp('A is not square matrix.')
  else
    B=A;
    z=zeros(n,n);
    L = tlist(['qd','d'],eye(n,n),z,z,z);
    U = tlist(['qd','d'],z,,z,z);
    for j=1:n
      for i=2:j
        for k=1:(i-1)
          B(i,j) = B(i,j) - B(i,k)*B(k,j);
        end
      end
      for i=(j+1):n
        for k=1:(j-1)
          B(i,j) = B(i,j) - B(i,k)*B(k,j);
       end
        B(i,j) = B(i,j) / B(j,j);
     end
    end
    for j=1:n
      U(1:j,j) = B(1:j,j);
      L((j+1):n,j) = B((j+1):n,j);
    end
  end
endfunction

