// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QD multiplication (double * qd)
//
// Calling Sequence :
// a * b
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function c = %s_m_qd(a,b)
  //D * QD
  [m1,n1] = size(a);
  [m2,n2] = size(b.d);
  //(1~1)~(1~1),(1~1)~(m~n),(m~n)~(1~1)
  if (m1 == 1 & n1 == 1) | (m2 == 1 & n2 == 1) then
    c = prod_qd_d(b,a)
  //(m~n)~(n~1)
  elseif n1 == m2 & n2 == 1 then
    A = a';B = b';
    c = prod_qd_d(B(1,1),A(1,:));
    for k=2:n1
      d = prod_qd_d(B(1,k),A(k,:));
      c = %qd_a_qd(c,d);
    end
    c = c';
  //(1~l)~(l~n)
  elseif m1 == 1 & n2 == m2 then
      A = a';B = b';
    c = prod_qd_d(B(:,1),A(1,1));
    for k=2:n1
      d = prod_qd_d(B(:,k),A(k,1));
      c = %qd_a_qd(c,d);
    end
    c = c';
  //(m~l)~(l~n)
  elseif n1 == m2 then
    A = a'; B=b';
    c = %qd_m_d(B,A);
    c = c';
  else
    error(10) //Inconsistent multiplication.
    abort;
  end
endfunction



function c = prod_qd_d(a,b)
  //QD multiplication
  //QD * D
  [p1,e1] = two_prod(a(2),b);
  [p2,e2] = two_prod(a(3),b);
  [p3,e3] = two_prod(a(4),b);
  p4 = a(5)*b;
  [p2,e1] = two_sum(p2,e1);
  [p3,e2,e1] = three_sum(p3,e2,e1);
  [p4,e2] = three_sum2(p4,e3,e2); 
  p5 = e1+e2;
  [t1,t2,t3,t4] = renormalize(p1,p2,p3,p4,p5);
  c = tlist(['qd','d'],t1,t2,t3,t4);
endfunction
